/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "operator/operator.pmh"

using namespace fe;

void HobbleOp::initialize(void)
{
	catalog<bool>("Translation Lock")=true;
	catalog<bool>("Translation Lock","joined")=true;
	catalog<String>("Translation Lock","hint")="Prevent input from shifting.";

	catalog<bool>("Rotation Lock")=true;
	catalog<String>("Rotation Lock","hint")="Prevent input from spinning.";

	catalog<bool>("ForceArms")=false;
	catalog<String>("ForceArms","label")="Force Indexes";
	catalog<bool>("ForceArms","joined")=true;
	catalog<String>("ForceArms","hint")=
			"Explicitly choose rotation control points.";

	catalog<I32>("ArmIndexX")= -1;
	catalog<I32>("ArmIndexX","min")= -1;
	catalog<I32>("ArmIndexX","max")=INT_MAX;
	catalog<String>("ArmIndexX","label")="Index To Side";
	catalog<bool>("ArmIndexX","joined")=true;
	catalog<String>("ArmIndexX","enabler")="ForceArms";
	catalog<String>("ArmIndexX","hint")=
		"Explicit point index towards one side of object,"
		" presumably well behaved (-1 to automatically choose).";

	catalog<I32>("ArmIndexZ")= -1;
	catalog<I32>("ArmIndexZ","min")= -1;
	catalog<I32>("ArmIndexZ","max")=INT_MAX;
	catalog<String>("ArmIndexZ","label")="Index To Front";
	catalog<String>("ArmIndexZ","enabler")="ForceArms";
	catalog<String>("ArmIndexZ","hint")=
		"Explicit point index towards front of object,"
		" presumably well behaved (-1 to automatically choose).";

	catalog<bool>("Force Source")=false;
	catalog<bool>("Force Source","joined")=true;
	catalog<String>("Force Source","hint")="Explicitly choose input center.";

	catalog<SpatialVector>("Source")=SpatialVector(0.0,0.0,0.0);
	catalog<String>("Source","enabler")="Force Source";
	catalog<String>("Source","hint")="Explicit center of input.";

	catalog<bool>("Force Destination")=false;
	catalog<bool>("Force Destination","joined")=true;
	catalog<String>("Force Destination","hint")="Explicitly place result.";

	catalog<SpatialVector>("Destination")=SpatialVector(0.0,0.0,0.0);
	catalog<String>("Destination","enabler")="Force Destination";
	catalog<String>("Destination","hint")="Explicit placement of result.";

	catalog<SpatialVector>("Input Center");
	catalog<String>("Input Center","IO")="output";
	catalog<String>("Input Center","hint")=
			"Approximate center of input surface.";

	catalog<SpatialVector>("Reference Center");
	catalog<String>("Reference Center","IO")="output";
	catalog<String>("Reference Center","hint")=
			"Approximate center of reference surface.";

	catalog<SpatialVector>("ConversionX");
	catalog<String>("ConversionX","label")="Rotation X";
	catalog<String>("ConversionX","IO")="output";

	catalog<SpatialVector>("ConversionY");
	catalog<String>("ConversionY","label")="Rotation Y";
	catalog<String>("ConversionY","IO")="output";

	catalog<SpatialVector>("ConversionZ");
	catalog<String>("ConversionZ","label")="Rotation Z";
	catalog<String>("ConversionZ","IO")="output";

	catalog<SpatialVector>("ConversionT");
	catalog<String>("ConversionT","label")="Translation";
	catalog<String>("ConversionT","IO")="output";

	catalog< sp<Component> >("Input Surface");

	catalog< sp<Component> >("Reference Surface");
	catalog<bool>("Reference Surface","optional")=true;
}

void HobbleOp::handle(Record& a_rSignal)
{
	SpatialVector& rInputCenter=catalog<SpatialVector>("Input Center");
	SpatialVector& rReferenceCenter=catalog<SpatialVector>("Reference Center");

	SpatialVector& rConversionX=catalog<SpatialVector>("ConversionX");
	SpatialVector& rConversionY=catalog<SpatialVector>("ConversionY");
	SpatialVector& rConversionZ=catalog<SpatialVector>("ConversionZ");
	SpatialVector& rConversionT=catalog<SpatialVector>("ConversionT");

	const bool forceArms=catalog<bool>("ForceArms");
	const I32 armIndexX=catalog<I32>("ArmIndexX");
	const I32 armIndexZ=catalog<I32>("ArmIndexZ");
	const bool forceSource=catalog<bool>("Force Source");
	const bool forceDestination=catalog<bool>("Force Destination");
	const SpatialVector source=catalog<SpatialVector>("Source");
	const SpatialVector destination=catalog<SpatialVector>("Destination");

	set(rInputCenter);
	set(rReferenceCenter);
	set(rConversionX,1.0,0.0,0.0);
	set(rConversionY,0.0,1.0,0.0);
	set(rConversionZ,0.0,0.0,1.0);
	set(rConversionT);

	sp<SurfaceAccessorI> spOutputPoint;
	if(!accessOutput(spOutputPoint,a_rSignal,e_point,e_position)) return;

	const U32 inputCount=spOutputPoint->count();
	if(!inputCount)
	{
		feLog("HobbleOp::handle point input count is zero\n");
		catalog<String>("error")="input points required";
		return;
	}

	sp<SurfaceAccessorI> spReferencePoint;
	access(spReferencePoint,"Reference Surface",e_point,e_position,e_quiet);

	U32 referenceCount=0;
	if(spReferencePoint.isValid())
	{
		referenceCount=spReferencePoint->count();
		if(referenceCount!=inputCount)
		{
			catalog<String>("warning")+=
					"reference point count different than input point count; ";
		}
	}

	const BWORD translationLock=catalog<bool>("Translation Lock");
	BWORD rotationLock=spReferencePoint.isValid() &&
			catalog<bool>("Rotation Lock");

	if(!translationLock && !rotationLock)
	{
		return;
	}

	Real indexX=0;
	Real indexZ=0;

	if(!forceArms || !forceSource || !forceDestination)
	{
		const U32 start=(forceArms && forceSource);
		const U32 passes=1+(spReferencePoint.isValid() &&
				(!forceArms || !forceDestination));
		for(U32 pass=start;pass<passes;pass++)
		{
			sp<SurfaceAccessorI> spPoint=pass? spReferencePoint: spOutputPoint;
			FEASSERT(spPoint.isValid());

			SpatialVector min=spPoint->spatialVector(0);
			SpatialVector max=min;
			for(U32 index=1;index<inputCount;index++)
			{
				const SpatialVector point=spPoint->spatialVector(index);

				for(U32 m=0;m<3;m++)
				{
					const Real value=point[m];
					if(min[m]>value)
					{
						min[m]=value;
					}
					if(max[m]<value)
					{
						max[m]=value;
						if(m==0)
						{
							indexX=index;
						}
						else if(m==2)
						{
							indexZ=index;
						}
					}
				}
			}

			if(pass)
			{
				rReferenceCenter=0.5*(min+max);
			}
			else
			{
				rInputCenter=0.5*(min+max);
			}
		}
	}

	const SpatialVector referenceTrue=rReferenceCenter;

	if(forceSource)
	{
		rInputCenter=source;
	}
	if(forceDestination)
	{
		rReferenceCenter=destination;
	}
	if(forceArms)
	{
		if(armIndexX>=0)
		{
			indexX=armIndexX;
		}
		if(armIndexZ>=0)
		{
			indexZ=armIndexZ;
		}
	}
	if(rotationLock)
	{
		if(indexX==indexZ)
		{
			catalog<String>("warning")+=
					"rotation control indexes can not match"
					" (rotation lock dismissed); ";
			rotationLock=FALSE;
		}
		if(indexX>=inputCount || indexZ>=inputCount)
		{
			catalog<String>("warning")+=
					"a rotation control index exceeds input point count"
					" (rotation lock dismissed); ";
			rotationLock=FALSE;
		}
		if(indexX>=referenceCount || indexZ>=referenceCount)
		{
			catalog<String>("warning")+=
					"a rotation control index exceeds reference point count"
					" (rotation lock dismissed); ";
			rotationLock=FALSE;
		}
	}

	const SpatialVector zero(0.0,0.0,0.0);
	SpatialTransform conversion;
	if(rotationLock)
	{
		//* rotation (translation optional)
		SpatialVector side=
				spReferencePoint->spatialVector(indexX)-referenceTrue;
		SpatialVector up=
				spReferencePoint->spatialVector(indexZ)-referenceTrue;
		SpatialTransform xformRef;
		BWORD success=makeFrameNormalY(xformRef,
				translationLock? rReferenceCenter: rInputCenter,
				side,unitSafe(up));

		side=spOutputPoint->spatialVector(indexX)-rInputCenter;
		up=spOutputPoint->spatialVector(indexZ)-rInputCenter;
		SpatialTransform xformDef;
		success=makeFrameNormalY(xformDef,rInputCenter,side,unitSafe(up));

		SpatialTransform invDef;
		invert(invDef,xformDef);
		conversion=invDef*xformRef;
	}
	else
	{
		//* translation only
		setIdentity(conversion);
		translate(conversion,rReferenceCenter-rInputCenter);
	}

	rConversionX=conversion.column(0);
	rConversionY=conversion.column(1);
	rConversionZ=conversion.column(2);
	rConversionT=conversion.column(3);

	for(U32 index=0;index<inputCount;index++)
	{
		const SpatialVector point=spOutputPoint->spatialVector(index);
		SpatialVector transformed;
		transformVector(conversion,point,transformed);

		spOutputPoint->set(index,transformed);
	}
}
